use async_trait::async_trait;
use figment::{
    providers::{Format, Serialized, Toml},
    util::map,
    value::Map,
    Figment,
};
use gethostname::gethostname;
use influxdb2::models::DataPoint;
use serde_derive::{Deserialize, Serialize};
use std::path::Path;
use tokio::sync::broadcast::*;

#[derive(Serialize, Deserialize, Clone)]
pub struct InfluxConfig {
    pub url: String,
    pub org: String,
    pub bucket: String,
    pub token: String,
    pub tags: Map<String, String>,
}

impl Default for InfluxConfig {
    fn default() -> Self {
        let host = gethostname().into_string().unwrap();
        InfluxConfig {
            url: "http://localhost:8086".into(),
            org: "default".into(),
            bucket: "default".into(),
            token: "".into(),
            tags: map! { "host".into() => host },
        }
    }
}

#[derive(Serialize, Deserialize, Clone)]
pub struct ChronyConfig {
    pub enabled: bool,
    pub timeout: u64,
    pub tracking_interval: u64,
    pub sources_interval: u64,
    pub measurement_prefix: String,
    pub tracking_measurement: String,
    pub sources_measurement: String,
    pub host: String,
}

impl Default for ChronyConfig {
    fn default() -> Self {
        ChronyConfig {
            enabled: false,
            timeout: 5,
            tracking_interval: 60,
            sources_interval: 300,
            measurement_prefix: "chrony.".into(),
            tracking_measurement: "tracking".into(),
            sources_measurement: "sources".into(),
            host: "127.0.0.1:323".into(),
        }
    }
}

#[derive(Serialize, Deserialize, Clone)]
pub struct HwmonSensorConfig {
    pub name: String,
    pub sensor: String,
}

#[derive(Serialize, Deserialize, Clone)]
pub struct HwmonConfig {
    pub enabled: bool,
    pub interval: u64,
    pub measurement_prefix: String,
    pub sensors: Map<String, HwmonSensorConfig>,
}

impl Default for HwmonConfig {
    fn default() -> Self {
        HwmonConfig {
            enabled: false,
            interval: 60,
            measurement_prefix: "hwmon.".into(),
            sensors: map! {},
        }
    }
}

#[derive(Serialize, Deserialize, Clone, Default)]
pub struct SourcesConfig {
    pub chrony: ChronyConfig,
    pub hwmon: HwmonConfig,
}

#[derive(Serialize, Deserialize, Clone, Default)]
pub struct Config {
    pub influxdb: InfluxConfig,
    pub sources: SourcesConfig,
}

pub fn load_config(filename: &Path) -> Figment {
    Figment::from(Serialized::defaults(Config::default())).merge(Toml::file(filename))
}

pub type ChimemonSourceChannel = Sender<DataPoint>;
pub type ChimemonTargetChannel = Receiver<DataPoint>;

#[async_trait]
pub trait ChimemonSource {
    async fn run(self, chan: ChimemonSourceChannel);
}

#[async_trait]
pub trait ChimemonTarget {
    async fn run(self, chan: ChimemonTargetChannel);
}
